#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include <getopt.h>
#include <omp.h>
#include "rabbits.h"

void usage() {
    fprintf(stderr,"Usage: %s -p <breed-probability> -t <time-steps>\n",
            program_invocation_short_name
           );
}

struct user_args get_user_args(const int argc,char **argv) {
    int opt;
    struct user_args args;

#ifdef DEBUG
        fprintf(stderr,"Processing user args\n");
#endif
    args.breed_ratio = args.time_steps = 0;
    while((opt = getopt(argc,argv,":p:t:")) != -1) {
        switch(opt) {
            case 'p':
                args.breed_ratio = strtof(optarg,NULL);
                if(args.breed_ratio == HUGE_VAL) {
                    fprintf(stderr,"Received invalid value: %s, %s:%d\n",
                            strerror(errno),__FILE__,__LINE__);
                    usage();
                    abort();
                }
                break;
            case 't':
                errno = 0;
                args.time_steps = (unsigned short int)strtol(optarg,NULL,10);
                if(errno) {
                    fprintf(stderr,"strtol failed: %s, %s:%d\n",
                            strerror(errno),__FILE__,__LINE__
                           );
                    abort();
                }
                break;
        }
    }

    if(args.breed_ratio == 0 || args.breed_ratio > 1
            || args.time_steps == 0) {
        fprintf(stderr,"Supply valid breed ratio!\n");
        usage();
        abort();
    }

#ifdef DEBUG
    fprintf(stderr,"Breed ratio = %f, time steps = %d\n",
            args.breed_ratio,args.time_steps
           );
#endif

    return args;
}

struct grassy_field *init_field(
        const struct user_args args,
        struct drand48_data *rand_buffer
        ) {

    struct grassy_field *field;
    size_t field_bytes = sizeof(grassy_field)*MAX_COORD;
    double rand_result;
    unsigned int i,x,y;

#ifdef DEBUG
    fprintf(stderr,"Initializing field of %d cells\n",MAX_COORD);
#endif

    if((field = malloc(field_bytes)) == NULL) {
        fprintf(stderr,"Failed to malloc %zd bytes for field: %s, %s:%d\n",
                field_bytes,strerror(errno),__FILE__,__LINE__
               );
        return NULL;
    }

    /* Set each cell to have no rabbit and maximum grass */
#ifdef OPENMP
#pragma omp parallel for collapse(2)
#endif
    for(x=0;x<MAX_X;x++) {
        for(y=0;y<MAX_Y;y++) {
            i = calc_2d21d(MAX_X,x,y);
            field[i].rabbit     = 0;
            field[i].grass      = MAX_GRASS;

            /* Cache neighbors if they are not outside our
             * range. Else set to OOB
             */

            field[i].neighbors.up 
                = y > 0 
                ? calc_2d21d(MAX_X,x,(y-1))
                : OOB;

            field[i].neighbors.left
                = x > 0
                ? calc_2d21d(MAX_X,(x-1),y)
                : OOB;

            field[i].neighbors.down
                = y < MAX_Y
                ? calc_2d21d(MAX_X,x,(y+1))
                : OOB;

            field[i].neighbors.right
                = x < MAX_X
                ? calc_2d21d(MAX_X,(x+1),y)
                : OOB;
        }
    }

    drand48_r(rand_buffer,&rand_result);
    i = (int)(rand_result*MAX_COORD);
#ifdef DEBUG
    fprintf(stderr,"Will start in %d\n",i);
#endif

    field[i].rabbit = MAX_HUNGER;

    return field;
}

unsigned int calc_2d21d(
        const unsigned int dim,
        const unsigned int x,
        const unsigned int y
        ) {
    return (dim*x)+y;
}

int breed_rabbits(
        const struct user_args args,
        struct grassy_field *field,
        struct drand48_data *rand_buffer
        ) {
    unsigned int i,t,x,y;

    for(t=0;t<args.time_steps;t++) {
#ifdef OPENMP
#pragma omp parallel for collapse(2) private(x,y,i) shared(field)
#endif
        for(x=0;x<MAX_X;x++) {
            for(y=0;y<MAX_Y;y++) {
                // Get the offset in the array from the x,y coords
                i = calc_2d21d(MAX_X,x,y);
               
                // Cell is empty, grow some grass!
                if(field[i].rabbit == DIE_HUNGER
                        && field[i].grass < MAX_GRASS) {
                    field[i].grass++;
#ifdef DEBUG
                    fprintf(stderr,"%d: Cell (%d,%d) now has %d grass\n",
                            t,x,y,field[i].grass
                           );
#endif
                }

                // Rabbit at 0 hunger dies
                if(field[i].rabbit > DIE_HUNGER) {
                    // Rabbit digests some food
                    field[i].rabbit--;
#ifdef DEBUG
                    fprintf(stderr,"%d: Rabbit at (%d,%d) digests, %d food left\n",
                            t,x,y,field[i].rabbit);
                    if(field[i].rabbit == DIE_HUNGER) {
                        fprintf(stderr,"Rabbit (%d,%d) dies\n",x,y);
                    }
#endif
                }

                // If we're still alive...
                if(field[i].rabbit > DIE_HUNGER) {
                    // Rabbits w/ food just stay and eat
                    if(field[i].grass > 0 && field[i].rabbit < MAX_HUNGER) {
                        // Two rabbit food points == one grass
                        field[i].grass--;
                        field[i].rabbit += 2;
#ifdef DEBUG
                        fprintf(stderr,"%d: Rabbit at (%d,%d) eats, %d grass left, %d food left\n",
                                t,x,y,field[i].grass,field[i].rabbit);
#endif
                    }
                    // Rabbits with at least MIN_BREED_HUNGER can spawn
                    if(field[i].rabbit >= MIN_BREED_HUNGER) {
#ifdef DEBUG
                        fprintf(stderr,"%d: Rabbit (%d,%d) is full enough to breed!\n",
                                t,x,y);
#endif
                        breed_rabbit(args,field,x,y,i,rand_buffer);
                    }
                    // Rabbits w/o food move
                    if (field[i].grass == 0) {
#ifdef OPENMP
#pragma omp critical
#endif
                        try_move(field,x,y,i,rand_buffer);
                    }
                }
            }
        }
        print_grid(field,t);
    }

    free(field);
    return 1;
}

bool breed_rabbit(
        const struct user_args args,
        struct grassy_field *field,
        const int unsigned x,
        const int unsigned y,
        const int unsigned i,
        struct drand48_data *rand_buffer
        ) {
    double rand_result;
    drand48_r(rand_buffer,&rand_result);
    int empty_cell;

    if(rand_result < args.breed_ratio) {
#ifdef DEBUG
        fprintf(stderr,"Rabbit %d (%d,%d) can breed!\n",i,x,y);
#endif
        empty_cell = rand_adj_empty_cell(field,i,rand_buffer);
        if(empty_cell == -1) {
#ifdef DEBUG
            fprintf(stderr,"No adjacent cells to breed into!\n");
#endif 
            return false;
        }
        field[empty_cell].rabbit = field[i].rabbit / 2;
        field[i].rabbit /= 2;
#ifdef DEBUG
        fprintf(stderr,"New rabbit (%d,%d), %d food\n",
                x,y,field[empty_cell].rabbit
               );
#endif
        return true;
    }
    return false;
}

unsigned short int adj_empty_cells(
        const struct grassy_field *field,
        const unsigned int i,
        unsigned int *empty_cell_list) {
    unsigned short int cells = 0;

    /* Look up, using cached neighbor information */
    if(field[i].neighbors.up != OOB
            && field[field[i].neighbors.up].rabbit 
                == DIE_HUNGER) {
        empty_cell_list[cells++] = field[i].neighbors.up;
    }

    /* Look down */
    if(field[i].neighbors.down != OOB
            && field[field[i].neighbors.down].rabbit
                == DIE_HUNGER) {
        empty_cell_list[cells++] = field[i].neighbors.down;
    }
    
    /* Look left */
    if(field[i].neighbors.left != OOB
            && field[field[i].neighbors.left].rabbit
                == DIE_HUNGER) {
        empty_cell_list[cells++] = field[i].neighbors.left;
    }

    /* Look right */
    if(field[i].neighbors.right != OOB
            && field[field[i].neighbors.right].rabbit
                == DIE_HUNGER) {
        empty_cell_list[cells++] = field[i].neighbors.right;
    }

#ifdef DEBUG
    fprintf(stderr,"Found %d empty cells in search\n",cells);
#endif
    return cells;
}

void print_grid(
        const struct grassy_field *field,
        const unsigned int t
        ) {
    unsigned int i,x,y,total_rabbits;

    total_rabbits = 0;
    for(x=0;x<MAX_X;x++) {
        for(y=0;y<MAX_Y;y++) {
            i = calc_2d21d(MAX_X,x,y);
            // printf("(%d,%d)\t",field[i].rabbit,field[i].grass);
            if(field[i].rabbit > DIE_HUNGER) {
                total_rabbits++;
            }
        }
        // puts("\n");
    }
    printf("\nTIMESTEP %d, total rabbits %d\n",
            t,total_rabbits);
}

void move_rabbit(
        struct grassy_field *field,
        const unsigned int old_cell,
        const unsigned int new_cell
        ) {
#ifdef DEBUG
    fprintf(stderr,"Moving rabbit from %d to %d\n",
            old_cell,new_cell
           );
#endif
    field[new_cell].rabbit = field[old_cell].rabbit;
    field[old_cell].rabbit = DIE_HUNGER;
}

int rand_adj_empty_cell(
        const struct grassy_field *field,
        const unsigned int i,
        struct drand48_data *rand_buffer
        ) {
    unsigned short int empty_cell_count;
    unsigned int *empty_cell_list,empty_cell;
    long int rand_result;

    // Maximum of four possibilities due to cardinal directions
    if((empty_cell_list = malloc(sizeof(int)*4)) == NULL) {
        fprintf(stderr,"Cannot allocate 4 ints for empty cell list: %s, %s:%d\n",
                strerror(errno),__FILE__,__LINE__
               );
        abort();
    }

    empty_cell_count = adj_empty_cells(field,i,empty_cell_list);

    if(empty_cell_count == 0) {
#ifdef DEBUG
        fprintf(stderr,"No empty cells surround (%d)\n",i);
#endif
        free(empty_cell_list);
        return -1;
    }
    else if(empty_cell_count == 1) {
        empty_cell = empty_cell_list[0];
    }
    // Need to make a random choice at this point
    else {
        lrand48_r(rand_buffer,&rand_result);
        empty_cell = empty_cell_list[(rand_result%empty_cell_count)];
    }

    free(empty_cell_list);

#ifdef DEBUG
    fprintf(stderr,"Random cell is %d\n",empty_cell);
#endif

    return empty_cell;
}

bool try_move(
        struct grassy_field *field,
        const unsigned int x,
        const unsigned int y,
        const unsigned int i,
        struct drand48_data *rand_buffer
        ) {
    unsigned int empty_cell;

#ifdef DEBUG
    fprintf(stderr,"No grass left, rabbit at (%d,%d) moves\n",x,y);
#endif
    empty_cell = rand_adj_empty_cell(field,i,rand_buffer);
    if(empty_cell == -1) {
#ifdef DEBUG
        fprintf(stderr,"No empty cells for rabbit at (%d,%d) to move to\n",
                x,y);
#endif
        return false;
    }
    move_rabbit(field,i,empty_cell);

    return true;
}
